﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using Tessa.Applications;
using Tessa.Applications.Messages;
using Tessa.Applications.Services.PlatformApplication;
using Tessa.Cards;
using Tessa.Extensions;
using Tessa.Extensions.Shared;
using Tessa.Platform;
using Tessa.Platform.Links;
using Tessa.Platform.Runtime;
using Tessa.Platform.Storage;
using Tessa.UI;
using Tessa.UI.AppManager.Synchronization;
using Tessa.UI.Client;
using Tessa.UI.Runtime;
using Tessa.UI.Windows;
using TessaDesktopApp.ViewModels;
using Unity;

namespace TessaDesktopApp
{
    /// <summary>
    /// Interaction logic for App.xaml
    /// </summary>
    public partial class App
    {
        #region Fields

        /// <summary>
        /// Шина для обработки сообщений от "AppManager".
        /// </summary>
        private IApplicationMessageBus bus;

        #endregion

        #region Base Overrides

        /// <inheritdoc/>
        protected override async Task OnApplicationStartupAsync(StartupEventArgs e, CancellationToken cancellationToken = default)
        {
            await TessaPlatform.InitializeFromConfigurationAsync(cancellationToken: cancellationToken);

            // подготовка UnityContainer-а
            this.UnityContainer = new UnityContainer();
            this.UnityContainer
                .RegisterClient()
                .RegisterApplicationsOnClient()
                .RegisterApplicationSynchronization();
            
            ClientRegistrator.ConfigureContainer(this.UnityContainer);
            this.UnityContainer
                .RegisterType<MainViewModel>(TypeLifetime.PerResolve)
                .RegisterApplicationServiceOnClient()
                .FindAndRegisterExtensionsOnClient(out List<string> actualFoldersList)
                .FinalizeClientRegistration(actualFoldersList);

            var linkManager = this.UnityContainer.Resolve<ILinkManager>();
            linkManager.Register(CardHelper.OpenCardLinkAction, this.OpenCardHandlerAsync);

            var resourceRegistrator = this.UnityContainer.Resolve<IApplicationResourceRegistrator>();
            resourceRegistrator.AddResourcesTo(this);

            var iconsRegistrator = this.UnityContainer.Resolve<IApplicationResourceRegistrator>(ApplicationResourceNames.Icons);
            iconsRegistrator.AddResourcesTo(UIHelper.Icons.MergedDictionaries);

            // запускаем приложение
            this.Application = this.UnityContainer.Resolve<IApplication>();

            this.Application.Launched += this.ApplicationLaunched;
            this.Application.ShuttingDown += this.ApplicationShuttingDown;

            // до начала запуска приложения используем значение из конфига
            ClientTracingHelper.RegisterExtensionTracing(this.UnityContainer);

            ApplicationLaunchResult result = await this.Application.LaunchAsync(
                new ApplicationLaunchParameters(
                    AbApplicationIdentifiers.TessaDesktopApp,
                    ct => this.Dispatcher.InvokeAsync(this.ActivateShell).Task,
                    e.Args)
                {
                    Splash = this.LauncherSplash,
                }, cancellationToken).ConfigureAwait(false);

            if (result != ApplicationLaunchResult.Launched)
            {
                return;
            }

            // создаём и отображаем главное окно
            Task task = await this.Dispatcher.InvokeAsync(async () =>
            {
                var viewModel = this.UnityContainer.Resolve<MainViewModel>();
                if (viewModel is IAsyncInitializable initializable)
                {
                    await initializable.InitializeAsync(cancellationToken);
                    // возвращаемся в поток UI
                }

                var session = this.UnityContainer.Resolve<ISession>();
                var applicationDescriptor = this.UnityContainer.Resolve<IApplicationDescriptor>();
                var configurationManager = this.UnityContainer.Resolve<IConfigurationManager>();
                var myTitleDescription = configurationManager.Configuration.Settings.TryGet<string>("TitleDescription");

                this.MainWindow = new TessaWindow
                {
                    Content = viewModel,
                    Title = string.Format("{0} ({1}){2} ({3}) - версия платформы {4} от {5}",
                        applicationDescriptor.Name,
                        session.ServerCode,
                        myTitleDescription is not null ? $", {myTitleDescription}" : string.Empty,
                        applicationDescriptor.ApplicationVersion,
                        BuildInfo.Version, 
                        FormattingHelper.FormatDate(BuildInfo.Date, convertToLocal: false)),
                    WindowStartupLocation = WindowStartupLocation.CenterScreen,
                    Icon = ApplicationHelper.ApplicationIcon,
                };

                this.MainWindow.Show();
            }).Task.ConfigureAwait(false);

            await task.ConfigureAwait(false);
        }

        #endregion

        #region Private Methods

        /// <summary>
        /// Обработчик события успешного запуска приложения.
        /// </summary>
        /// <param name="sender">Объект, отправитель события.</param>
        /// <param name="e">Аргументы события события.</param>
        private async void ApplicationLaunched(object sender, ApplicationContextDeferredEventArgs e)
        {
            if (e.Context == null || e.Context.Parameters == null || !e.Context.Parameters.LaunchedByAppManager)
            {
                return;
            }

            this.bus = this.UnityContainer.Resolve<IApplicationMessageBus>();

            using (e.Defer())
            {
                await this.bus.RegisterAsync(ApplicationApiVersionNames.Get(e.Context.Parameters.AppManagerApi)).ConfigureAwait(false);
                this.bus.MessageReceived += this.OnMessageReceived;
            }
        }

        /// <summary>
        /// Обработчик события завершения приложения.
        /// </summary>
        /// <param name="sender">Объект, отправитель события.</param>
        /// <param name="e">Аргументы события события.</param>
        private void ApplicationShuttingDown(object sender, ApplicationContextDeferredEventArgs e)
        {
            if (e.Context == null || e.Context.Parameters == null || !e.Context.Parameters.LaunchedByAppManager || this.bus == null)
            {
                return;
            }

            this.bus.Dispose();
            this.bus = null;
        }

        /// <summary>
        /// Обработчик поступления сообщения от AppManager-а.
        /// </summary>
        /// <param name="sender">Объект, отправитель события.</param>
        /// <param name="e">Аргументы события события.</param>
        private async void OnMessageReceived(object sender, MessageReceivedEventArgs e)
        {
            if (e.Message is ApplicationLinkMessage linkMessage)
            {
                e.Handled = true;

                var linkManager = this.UnityContainer.Resolve<ILinkManager>();
                var messageProvider = this.UnityContainer.Resolve<IMessageProvider>();
                Func<CancellationToken, Task> activateShellActionAsync = ct => this.Dispatcher.InvokeAsync(this.ActivateShell).Task;

                bool handled = await linkManager.ProcessLinkAsync(
                    linkMessage.LinkArguments,
                    messageProvider,
                    this.UnityContainer,
                    activateShellActionAsync).ConfigureAwait(false);

                if (!handled)
                {
                    // пришло сообщение от AppManager, а приложение не смогло распарсить ссылку -
                    // всё равно отображаем окно приложения
                    await activateShellActionAsync(CancellationToken.None).ConfigureAwait(false);
                }
            }
        }

        /// <summary>
        /// Обработчик события открытия карточки по ссылке от AppManager-а.
        /// </summary>
        /// <param name="context">Контекст ссылки на открытие карточки.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task OpenCardHandlerAsync(ILinkContext context)
        {
            if (!CardHelper.TryParseLink(
                context.Parameters,
                out Guid cardID,
                out string displayValue,
                out Guid? fileID,
                out Guid? versionID))
            {
                return;
            }

            // если окно было свёрнуто - разворачиваем его
            await context.ActivateShellAsync(context.CancellationToken).ConfigureAwait(false);

            Task task = await this.Dispatcher.InvokeAsync(async () =>
            {
                var viewModel = this.MainWindow.Content as MainViewModel;
                if (viewModel is null)
                {
                    return;
                }
                viewModel.TrySelectCard(cardID);
            }).Task.ConfigureAwait(false);

            await task.ConfigureAwait(false);
            context.Handled = true;
        }

        #endregion
    }
}
