﻿using System;
using System.Threading;
using System.Threading.Tasks;
using Tessa.Platform;
using Tessa.UI;
using Tessa.UI.Cards;
using Tessa.UI.Cards.Controls;
using Tessa.UI.Views.Content;
using Tessa.Views.Metadata;

namespace TessaDesktopApp.ViewModels
{
    /// <summary>
    /// Главная модель представления.
    /// </summary>
    public sealed class MainViewModel :
        ViewModel<EmptyModel>,
        IAsyncInitializable
    {
        #region Fields

        /// <summary>
        /// Делегат создания модели типа диалога.
        /// </summary>
        private readonly CreateDialogFormFuncAsync createDialogFormFuncAsync;

        /// <summary>
        /// Функция-помощник создания редактора.
        /// </summary>
        private readonly Func<ICardEditorModel> createEditorFunc;

        /// <summary>
        /// Модель представления списка доступных карточек.
        /// </summary>
        private IFormViewModel viewFormViewModel;

        /// <summary>
        /// Модель представления для элемента управления "Представление".
        /// </summary>
        private CardViewControlViewModel viewControlViewModel;

        /// <summary>
        /// Модель представления редактора карточек.
        /// </summary>
        private ICardEditorModel cardEditorViewModel;

        /// <summary>
        /// Псевдоним колонки с идентификатором карточки.
        /// </summary>
        private string cardIdName;

        #endregion

        #region Constructors

        /// <summary>
        /// Конструктор главной модели представления с разрешением зависимостей из Unity.
        /// </summary>
        /// <param name="createDialogFormFuncAsync">Делегат создания модели типа диалога.</param>
        /// <param name="createEditorFunc">Функция создания редактора карточек.</param>
        public MainViewModel(CreateDialogFormFuncAsync createDialogFormFuncAsync,
            Func<ICardEditorModel> createEditorFunc)
        {
            this.createDialogFormFuncAsync = createDialogFormFuncAsync;
            this.createEditorFunc = createEditorFunc;
        }

        #endregion

        #region ViewModel Properties

        /// <summary>
        /// Модель представления списка доступных карточек.
        /// </summary>
        public IFormViewModel ViewFormViewModel => this.viewFormViewModel;

        /// <summary>
        /// Модель представления редактора карточек.
        /// </summary>
        public ICardEditorModel CardEditorViewModel => this.cardEditorViewModel;

        #endregion

        #region IAsyncInitializable

        /// <inheritdoc/>
        async ValueTask IAsyncInitializable.InitializeAsync(CancellationToken cancellationToken)
        {
            // получаем модель представления со списком доступных карточек
            ICardModel cardModel;
            (this.viewFormViewModel, cardModel) = await this.createDialogFormFuncAsync(
                "AbUserInterface", "View", cancellationToken: cancellationToken);

            // получаем и настраиваем редактор карточки
            this.cardEditorViewModel = this.createEditorFunc();
            this.cardEditorViewModel.DialogName = CardUIHelper.DefaultDialogName;
            // настраиваем обработчик для получения сообщений об изменении выделенной строки
            this.viewControlViewModel = cardModel.Controls.TryGet<CardViewControlViewModel>("ViewControl");
            if (this.viewControlViewModel is null)
            {
                return;
            }
            this.cardIdName = TryGetCardIDName(this.viewControlViewModel.ViewMetadata);
            this.viewControlViewModel.Table.RowSelected += async (o, e) =>
            {
                await this.SelectViewRowAsync(e.Row);
            };

            await this.SelectViewRowAsync(this.viewControlViewModel.Table.SelectedItem, cancellationToken);
        }

        #endregion

        #region Internal Methods

        /// <summary>
        /// Выполняет попытку открыть карточку с заданным идентификатором.
        /// </summary>
        /// <param name="cardID">Идентификатор карточки.</param>
        internal void TrySelectCard(Guid cardID)
        {
            var table = this.viewControlViewModel?.Table;
            if (table is null || string.IsNullOrEmpty(this.cardIdName))
            {
                return;
            }
            foreach (var row in table.Items)
            {
                if (row.Data[this.cardIdName] is not Guid ID) 
                {
                    continue;
                }
                if (ID == cardID)
                {
                    table.SelectedItem = row;
                    break;
                }
            }
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Обработчик события выделения элемента в представлении.
        /// </summary>
        /// <param name="row">Выделенная строка.</param>
        /// <param name="cancellationToken">Объект для отмены асинхронной операции.</param>
        /// <returns>Асинхронная задача.</returns>
        private async Task SelectViewRowAsync(
            TableRowViewModel row, 
            CancellationToken cancellationToken = default)
        {
            // нет идентификатора карточки, синхронизация невозможна
            if (string.IsNullOrEmpty(this.cardIdName) || row is null)
            {
                return;
            }
            if (row.Data[this.cardIdName] is not Guid cardID ||
                cardID == Guid.Empty)
            {
                return;
            }

            // открываем выделенную карточку
            await using (UIContext.Create(this.cardEditorViewModel.Context))
            {
                await this.cardEditorViewModel.OpenCardAsync(cardID, null, null, 
                    context: this.cardEditorViewModel.Context,
                    cancellationToken: cancellationToken);
            }
        }

        /// <summary>
        /// Помощник получения псевдонима колонки с идентификатором карточки из представления.
        /// </summary>
        /// <param name="viewMetadata">Метаданные представления.</param>
        /// <returns>Возвращает строку с именем колонки идентификатора карточки, или <c>null</c>, если её не удалось найти.</returns>
        private static string TryGetCardIDName(IViewMetadata viewMetadata)
        {
            if (viewMetadata is null)
            {
                return null;
            }

            foreach (var reference in viewMetadata.References)
            {
                if (!reference.IsCard || !reference.OpenOnDoubleClick)
                {
                    continue;
                }
                var column = viewMetadata.Columns.FindByName(reference.ColPrefix + "ID") ??
                    viewMetadata.Columns.FindByName(reference.ColPrefix + "RowID");
                if (column is not null)
                {
                    return column.Alias;
                }
            }

            return null;
        }

        #endregion
    }
}
